@@ -1,5 +1,5 @@
 /* Copyright (c) 2011 Nanakos Chrysostomos <nanakos@wired-net.gr>
-   Simple and lightweight Yubikey OTP-OAUTH/HOTP Validation Server
+   Simple and lightweight Yubikey OTP-OATH/HOTP Validation Server
 
    yubiserver is placed under the GNU General Public License, version 2 or later.
 
@@ -30,17 +30,20 @@
 #include <gcrypt.h>
 #include <sqlite3.h>
 #include <time.h>
+#include <sys/timeb.h>
 #include <getopt.h>
 #include <math.h>
 #include <mhash.h>
 
+#include <ev.h>
+
 #include <libconfig.h>
 #include "yubiserver.h"
 
 void yubilog(int type, const char *s1, const char *s2, int num)
 {
 	int fd ;
-	char logbuffer[BUFSIZE*2];
+	char *logbuffer=malloc(BUFSIZE*2);
 
 	switch (type) 
 	{
@@ -63,6 +66,7 @@ void yubilog(int type, const char *s1, const char *s2, int num)
 		(void)write(fd,"\n",1);      
 		(void)close(fd);
 	}
+	free(logbuffer);
 	if(type == ERROR || type == WARNING) 
 		exit(3);
 }
@@ -79,26 +83,28 @@ void null_terminate(char *buffer)
 }
 
 /* Decode Modhex to Hex */
-char *modhex2hex(char *otp)
+void modhex2hex(char *otp_n,char *otp)
 {
 	char *hex="0123456789abcdef";
 	char *modhex="cbdefghijklnrtuv";
-	char *otp_n;
 	int i;
 	char *end;
 	int pos;
-	otp_n=(char *)malloc((size_t)OTP_TOKEN);
 	for(i=0;i<strlen(otp);i++)
 	{
 		end=index(modhex,otp[i]);
+		if(end==NULL) goto out;
 		pos=end-modhex;
 		otp_n[i]=hex[pos];
 	}
-	return otp_n;
+	return;
+out:
+	memcpy(otp_n,otp,strlen(otp));
+	return;
 }
 
-/* Convert ASCII to Hexadecimal  for OAUTH*/
-void oauth_atoh(char *token,char *out)
+/* Convert ASCII to Hexadecimal  for OATH*/
+void oath_atoh(char *token,char *out)
 {
 	int temp,index;
 	char buf[2]={' ','\0'};
@@ -154,7 +160,7 @@ int get_aeskey(char *otp,char *aeskey,char *private_id)
 	sqlite3 *handle;
 	sqlite3_stmt *stmt;
 	int retval,rows=0;
-	char * query = (char *)malloc((size_t)100);
+	char * query = (char *)malloc((size_t)QUERY_SIZE);
 	char * public_id = (char *)malloc((size_t)PRIVATE_ID_SIZE+1);
 	/* Get public_id*/
 	get_publicid(otp,public_id);
@@ -189,6 +195,7 @@ int get_aeskey(char *otp,char *aeskey,char *private_id)
 			break;
 		}
 	}
+	sqlite3_finalize(stmt);
 	sqlite3_close(handle);	
 	free(query);
 	free(public_id);
@@ -202,8 +209,8 @@ int get_aeskey(char *otp,char *aeskey,char *private_id)
 /* Decrypt AES encrypted OTP */
 char *aes128ecb_decrypt(char *otp,char *premod_otp,char *private_id)
 {
-	#define GCRY_CIPHER GCRY_CIPHER_AES128   // Pick the cipher
-	#define GCRY_C_MODE GCRY_CIPHER_MODE_ECB // Pick the cipher mode
+	#define GCRY_CIPHER GCRY_CIPHER_AES128   // cipher
+	#define GCRY_C_MODE GCRY_CIPHER_MODE_ECB // cipher mode
 	gcry_error_t     gcryError;
     	gcry_cipher_hd_t gcryCipherHd;
 	size_t           index;
@@ -214,41 +221,65 @@ char *aes128ecb_decrypt(char *otp,char *premod_otp,char *private_id)
 	char * otp_buffer = (char *)malloc((size_t)32);
 	char * final_otp = (char *)malloc(33);
 	char * aesSymKey = (char *)malloc(32);
-	int retval=get_aeskey(premod_otp,aesSymKey,private_id);
-
+	gcry_control (GCRYCTL_ANY_INITIALIZATION_P);
+	//gcry_check_version ( NULL );
+        gcry_control ( GCRYCTL_DISABLE_SECMEM_WARN );
+        gcry_control ( GCRYCTL_INIT_SECMEM, 16384, 0 );
+	gcry_control (GCRYCTL_INITIALIZATION_FINISHED, NULL,0);
+	int retval = get_aeskey(premod_otp,aesSymKey,private_id);
 	if(retval<0)
 	{
+		free(decoded_otp);
+		free(otp_token);
+		free(otp_buffer);
+		free(final_otp);
+		free(aesSymKey);
 		return NULL;
 	}
 
 	char * aesKey = (char *)malloc((size_t)16);
+
 	gcryError = gcry_cipher_open(
 			&gcryCipherHd, // gcry_cipher_hd_t *
 			GCRY_CIPHER,   // int
 			GCRY_C_MODE,   // int
 			0);            // unsigned int
 
-	if (gcryError)
+	if(gcryError)
 	{
 		//yubilog(ERROR,"gcry_cipher_open failed",gcry_strsource(gcryError),gcry_strerror(gcryError));
 		yubilog(LOG,"gcry_cipher_open failed",gcry_strsource(gcryError),0);
+		free(decoded_otp);
+		free(otp_token);
+		free(otp_buffer);
+		free(final_otp);
+		free(aesSymKey);
+		free(aesKey);
 		return NULL;
 	}
 	atoh(aesSymKey,aesKey);
 	gcryError = gcry_cipher_setkey(gcryCipherHd, aesKey, keyLength);
 
-	if (gcryError)
+	if(gcryError)
 	{
 		//yubilog(ERROR,"gcry_cipher_setkey failed",gcry_strsource(gcryError),gcry_strerror(gcryError));
 		yubilog(LOG,"gcry_cipher_setkey failed",gcry_strsource(gcryError),0);
+		free(decoded_otp);
+		free(otp_token);
+		free(otp_buffer);
+		free(final_otp);
+		free(aesSymKey);
+		free(aesKey);
+		gcry_cipher_close(gcryCipherHd);
 		return NULL;
 	}
 
-	for(index=0;index<32;index++)
+	memcpy(otp_buffer,otp+12,32);
+/*	for(index=0;index<32;index++)
 	{
 		otp_buffer[index]=otp[index+12];
 	}
-
+*/
 	atoh(otp_buffer,otp_token);
 	gcryError = gcry_cipher_decrypt(
 			gcryCipherHd, // gcry_cipher_hd_t
@@ -256,19 +287,27 @@ char *aes128ecb_decrypt(char *otp,char *premod_otp,char *private_id)
 			otpLength,    // size_t
 			otp_token,    // const void *
 			keyLength);   // size_t
-	if (gcryError)
+	if(gcryError)
 	{
 		//yubilog(ERROR,"gcry_cipher_decrypt failed",gcry_strsource(gcryError),gcry_strerror(gcryError));
 		yubilog(LOG,"gcry_cipher_decrypt failed",gcry_strsource(gcryError),0);
+		free(decoded_otp);
+		free(otp_token);
+		free(otp_buffer);
+		free(final_otp);
+		free(aesSymKey);
+		free(aesKey);
+		gcry_cipher_close(gcryCipherHd);
 		return NULL;
 	}
-
+	gcry_cipher_reset(gcryCipherHd);
 	gcry_cipher_close(gcryCipherHd);
 	htoa(decoded_otp,final_otp);
 	free(decoded_otp);
 	free(otp_token);
 	free(otp_buffer);
 	free(aesSymKey);
+	free(aesKey);
 	return final_otp;
 }
 
@@ -321,6 +360,7 @@ char *base64_encode(const unsigned char *data,size_t input_length,size_t output_
 
         for (i = 0; i < mod_table[input_length % 3]; i++)
                 encoded_data[output_length - 1 - i] = '=';
+	encoded_data[output_length]='\0';
         return encoded_data;
 }
 
@@ -392,7 +432,7 @@ int sqlite3_countertimestamp(char *otp,int *db_counter,int *db_timestamp,struct
                         *db_counter = atoi((const char *)sqlite3_column_text(stmt,0));
                         *db_timestamp = atoi((const char *)sqlite3_column_text(stmt,1));
 			sprintf(yubikey->creation_date,"%s",sqlite3_column_text(stmt,2));
-			yubikey->creation_date[20]='\0';
+			yubikey->creation_date[24]='\0';
                         rows++;
                 }
                 else if(retval==SQLITE_DONE)
@@ -405,6 +445,7 @@ int sqlite3_countertimestamp(char *otp,int *db_counter,int *db_timestamp,struct
 			break;
                 }
         }
+	sqlite3_finalize(stmt);
         sqlite3_close(handle);
         free(query);
         free(public_id);
@@ -430,12 +471,14 @@ void sqlite3_updatecounter(char *otp,int counter,int timestamp)
         retval = sqlite3_open(sqlite3_dbpath,&handle);
         
 	if(retval)
-                yubilog(ERROR,"sqlite3_updatecounter: Database connection failed",0,0);
+                yubilog(LOG,"sqlite3_updatecounter: Database connection failed",0,0);
 
         retval = sqlite3_exec(handle,query,0,0,0);
         
-	if(retval)
-                yubilog(ERROR,"sqlite3_updatecounter:Updating counter/timestamp data to DB failed",0,0);
+	if(retval != SQLITE_OK)
+	{
+                yubilog(LOG,"sqlite3_updatecounter:Updating counter/timestamp data to DB failed",0,0);
+	}
         
 	sqlite3_close(handle);
         
@@ -448,14 +491,14 @@ char *get_apikey(char *id)
 	sqlite3 *handle;
 	sqlite3_stmt *stmt;
 	int retval,rows=0;
-	char * query = (char *)malloc((size_t)100);
+	char * query = (char *)malloc((size_t)QUERY_SIZE);
 	char * secret = (char *)malloc((size_t)40);
 	/* Create query for aeskey and private id */
 	sprintf(query,"SELECT secret FROM apikeys WHERE id='%s'",id);
 	retval = sqlite3_open(sqlite3_dbpath,&handle);
 	
 	if(retval)
-		yubilog(ERROR,"get_apikey: Database connection failed",0,0);
+		yubilog(LOG,"get_apikey: Database connection failed",0,0);
 		
 	retval = sqlite3_prepare_v2(handle,query,-1,&stmt,0);
 	
@@ -481,13 +524,17 @@ char *get_apikey(char *id)
 		}
 	}
 
+	sqlite3_finalize(stmt);
 	sqlite3_close(handle);
 	free(query);
 
 	if(rows)
 		return secret; /* If user found return secret */
 	else
+	{
+		free(secret);
 		return NULL; /* If the user does not belong to the DB or is deactivated return NULL value */
+	}
 }
 
 /* Validate OTP */
@@ -501,6 +548,7 @@ int validate_otp(char *otp,char *premod_otp,char *private_id,struct Yubikey *yub
 
 	if(strncmp(otp,private_id,PRIVATE_ID_SIZE)!=0)
 	{	
+		free(bcrc);
 		return BAD_OTP;
 	}
 
@@ -555,10 +603,13 @@ char *create_hmac(char *otp,char *status,char *datetime,char *id,char *nonce)
 	
 	char *password = get_apikey(id);
 
-	if(password == NULL) return NULL;
+	if(password == NULL)
+	{
+		 return NULL;
+	}
 
 	keylen = strlen(password);
-	data = (char *)malloc((char)350);
+	data = (char *)calloc(1,180);
 	if(nonce!=NULL)
 	{
 		datalen = sprintf(data,"nonce=%s&otp=%s&sl=100&status=%s&t=%s",nonce,otp,status,datetime);
@@ -571,7 +622,8 @@ char *create_hmac(char *otp,char *status,char *datetime,char *id,char *nonce)
 	mhash(td,data,datalen);
 	mhash_hmac_deinit(td,mac);
 	char *output = base64_encode(mac,20,20);
-
+	free(data);
+	free(password);
 	return output;
 }
 
@@ -581,6 +633,7 @@ char *hotp(char *key,long counter,int digits)
 	int bin_code;
 	unsigned char hmac_result[20];
 	int offset;
+	char *final_hotp = NULL;
 	char Counter[8]={
 		((long)counter >> 56) & 0xff,((long)counter >> 48) & 0xff,((long)counter >> 40) & 0xff,
 		((long)counter >> 32) & 0xff,((long)counter >> 24) & 0xff,((long)counter >> 16) & 0xff,
@@ -595,29 +648,32 @@ char *hotp(char *key,long counter,int digits)
 		| (hmac_result[offset+2] & 0xff) <<  8
 		| (hmac_result[offset+3] & 0xff) ;
 	sprintf(HOTP,"%d",bin_code);
-
-	return (HOTP+strlen(HOTP)-digits); /* Digits is usually 6 */
+	final_hotp = calloc(1,digits+1);
+	memcpy(final_hotp,HOTP+strlen(HOTP)-digits,digits); /* Digits is usually 6 */
+	free(HOTP);
+	return final_hotp;
+//	return (HOTP+strlen(HOTP)-digits); /* Digits is usually 6 */
 }
 
 
-struct Yubikey *oauth_counter_secret(char *id)
+struct Yubikey *oath_counter_secret(char *id)
 {
 	sqlite3 *handle;
 	sqlite3_stmt *stmt;
 	int retval,rows=0;
-	char * query = (char *)malloc((size_t)100);
+	char * query = (char *)malloc((size_t)QUERY_SIZE);
 	struct Yubikey *oyubikey = (struct Yubikey *)calloc(1,sizeof(struct Yubikey));
 	
 	sprintf(query,"SELECT counter,secret FROM oathtokens WHERE publicname='%.12s' AND active='1'",id);
 	retval = sqlite3_open(sqlite3_dbpath,&handle);
 	
 	if(retval)
-		yubilog(ERROR,"oauth_counter_secret: Database connection failed",0,0);
+		yubilog(LOG,"oath_counter_secret: Database connection failed",0,0);
 		
 	retval = sqlite3_prepare_v2(handle,query,-1,&stmt,0);
 	
 	if(retval)
-		yubilog(LOG,"oauth_counter_secret: Selecting data from DB failed",0,0);
+		yubilog(LOG,"oath_counter_secret: Selecting data from DB failed",0,0);
 	
 	while(1)
 	{
@@ -634,11 +690,12 @@ struct Yubikey *oauth_counter_secret(char *id)
 		}
 		else
 		{
-			yubilog(LOG,"oauth_counter_secret: Database error encountered",0,0);
+			yubilog(LOG,"oath_counter_secret: Database error encountered",0,0);
 			break;
 		}
 	}
 
+	sqlite3_finalize(stmt);
 	sqlite3_close(handle);
 	free(query);
 
@@ -651,7 +708,7 @@ struct Yubikey *oauth_counter_secret(char *id)
 	}
 }
 
-void sqlite3_oauth_updatecounter(char *otp,int counter)
+void sqlite3_oath_updatecounter(char *otp,int counter)
 {
 	sqlite3 *handle;
         int retval;
@@ -665,50 +722,66 @@ void sqlite3_oauth_updatecounter(char *otp,int counter)
         retval = sqlite3_open(sqlite3_dbpath,&handle);
         
 	if(retval)
-                yubilog(ERROR,"sqlite3_oauth_updatecounter: Database connection failed",0,0);
+                yubilog(LOG,"sqlite3_oath_updatecounter: Database connection failed",0,0);
 
         retval = sqlite3_exec(handle,query,0,0,0);
         
 	if(retval)
-                yubilog(ERROR,"sqlite3_oauth_updatecounter:Updating counter/timestamp data to DB failed",0,0);
+                yubilog(LOG,"sqlite3_oath_updatecounter:Updating counter/timestamp data to DB failed",0,0);
+	
+	free(query);
+        free(public_id);
         
 	sqlite3_close(handle);
         
-	free(query);
-        free(public_id);
 }
 
-/* Validate HOTP/OAUTH */
-int validate_hotp(char *id,char *otp,struct OAUTH_Tokens *tokens)
+/* Validate HOTP/OATH */
+int validate_hotp(char *id,char *otp,struct OATH_Tokens *tokens)
 {
 
 	char *tokenid = (char *)malloc((size_t)PRIVATE_ID_SIZE+1);
 	char *hotp_val = NULL;
 	char *Key = NULL;
+	char *temp = NULL;
 	int counter;
 	get_publicid(otp,tokenid);
 	tokenid[PRIVATE_ID_SIZE]='\0';
-	struct Yubikey *oyubikey = oauth_counter_secret(tokenid);
+	struct Yubikey *oyubikey = oath_counter_secret(tokenid);
 
-	if(oyubikey == NULL) return BAD_OTP;
+	if(oyubikey == NULL) 
+	{
+		free(tokenid);
+		return BAD_OTP;
+	}
 	
 	hotp_val = (char *)malloc(strlen(otp)-12+1);
 	sprintf(hotp_val,"%s",otp+12);
-	if(strlen(hotp_val) % 2 != 0) return BAD_OTP;
+	if(strlen(hotp_val) % 2 != 0) 
+	{	
+		free(tokenid);
+		free(hotp_val);
+		free(oyubikey);
+		return BAD_OTP;
+	}
 	Key = (char *)malloc((size_t)21);
-	oauth_atoh(oyubikey->private_id,Key);
+	oath_atoh(oyubikey->private_id,Key);
 	Key[20]='\0';
 	for(counter=1+oyubikey->counter ; counter < 256+oyubikey->counter; counter++)
 	{
-		if(!strcmp(hotp_val,hotp(Key,counter,strlen(hotp_val))))
+		temp=hotp(Key,counter,strlen(hotp_val)); 
+		//if(!strcmp(hotp_val,hotp(Key,counter,strlen(hotp_val))))
+		if(!strcmp(hotp_val,temp))
 		{
 			tokens->counter = counter;
 			if(Key) free(Key);
+			if(temp) free(temp);
 			if(oyubikey) free(oyubikey);
 			if(hotp_val) free(hotp_val);
 			free(tokenid);
 			return OK;
 		}
+		free(temp);
 	}
 	if(Key) free(Key);
 	if(oyubikey) free(oyubikey);
@@ -716,15 +789,15 @@ int validate_hotp(char *id,char *otp,struct OAUTH_Tokens *tokens)
 	free(tokenid);
 	return NO_AUTH;
 }
-/* Validate OTP-HOTP/OAUTH */
+/* Validate OTP-HOTP/OATH */
 int validate_all(char *otp,char *premod_otp,char *private_id,char *id,struct Yubikey *yubikey,int method,char **addr)
 {
 	int retval=BAD_OTP;
 	if(method == METHOD_OTP)
 		retval = validate_otp(otp,premod_otp,private_id,yubikey);
-	if(method == METHOD_OAUTH)
+	if(method == METHOD_OATH)
 	{
-		retval = validate_hotp(id,otp,container_of(addr,struct OAUTH_Tokens,otp));
+		retval = validate_hotp(id,otp,container_of(addr,struct OATH_Tokens,otp));
 	}
 	return retval;
 }
@@ -770,13 +843,12 @@ struct Tokens *tokenize(char *buffer)
 		if(strstr(token[j],"sl=")!= NULL) tokens->sl = find_token(token[j]);
 		if(strstr(token[j],"timeout=")!= NULL) tokens->timeout = atoi(find_token(token[j]));
 	}
-	
 	return tokens;
 }
 
-struct OAUTH_Tokens *oauth_tokenize(char *buffer)
+struct OATH_Tokens *oath_tokenize(char *buffer)
 {
-	struct OAUTH_Tokens *tokens = (struct OAUTH_Tokens *)calloc(1,sizeof(struct OAUTH_Tokens));
+	struct OATH_Tokens *tokens = (struct OATH_Tokens *)calloc(1,sizeof(struct OATH_Tokens));
 	char *token1,*token2,*token[7];
 	int j,i=0;
 	tokens->id = NULL;
@@ -802,186 +874,293 @@ struct OAUTH_Tokens *oauth_tokenize(char *buffer)
 }
 
 /* Child authentication server process, exit on errors */
-void authentication_handle(int fd, int requestno)
+static void write_callback(struct ev_loop *loop, struct ev_io *w, int revents)
 {
 	long i, ret;
 	char *fstr = NULL;
-	char *otp = NULL, *id = NULL, *nonce = NULL;
-	char validation_date[20];
+	char *otp = NULL;
 	char *hmac = NULL;
+	char *private_id = NULL;
+	struct Tokens  *tokens = NULL;
+	struct OATH_Tokens  *oath_tokens = NULL;
+	static char validation_date[24];
+	static char datetmp[20];
 	time_t t;
 	struct tm *tmp = NULL;
-	static char buffer[BUFSIZE+1]; /* static so zero filled */
+	struct timeb tp;
+	char *buffer = (char *)calloc(1,BUFSIZE+1); /* zero filled */
 	int result = BAD_OTP;
-	char *premod_otp = NULL;
-	struct Tokens  *tokens = NULL;
-	struct OAUTH_Tokens  *oauth_tokens = NULL;
 	int sessioncounter=0,sessionuse=0,session_ts=0;
 	int sl = -1;
-	char *private_id = (char *)malloc((size_t)PRIVATE_ID_SIZE);
 	char *status[12]={"OK","BAD_OTP","REPLAYED_OTP","DELAYED_OTP","NO_SUCH_CLIENT","BAD_SIGNATURE","MISSING_PARAMETER", 
 		          "OPERATION_NOT_ALLOWED","BACKEND_ERROR","NOT_ENOUGH_ANSWERS","REPLAYED_REQUEST","NO_AUTH"
 			 }; 
 	struct Yubikey *yubikey = NULL;
+	struct ev_client *cli= ((struct ev_client*) (((char*)w) - offsetof(struct ev_client,ev_write)));
+	memcpy(buffer,cli->buffer,cli->ret);
+	ret = cli->ret;
+	
 	//struct Yubikey *yubikey = (struct Yubikey *)calloc(1,sizeof(struct Yubikey));
 	//struct Yubikey *yubikey = (struct Yubikey *)malloc(sizeof(struct Yubikey));	
 	//yubikey->creation_date[0]='\0';
-	ret =read(fd,buffer,BUFSIZE); 	/* read server request in one go */
 	
 	if(ret == 0 || ret == -1) 	/* read failure then stop now */
-		yubilog(WARNING,"failed to read browser request","",fd);
+	{
+		if (revents & EV_WRITE)
+			ev_io_stop(EV_A_ w);
+		yubilog(LOG,"failed to read browser request",buffer,ret);
+		goto read_error;
+	}
 	
 	if(ret > 0 && ret < BUFSIZE)	/* return code is valid chars */
 		buffer[ret]=0;		/* terminate the buffer */
-	else buffer[0]=0;
+	else buffer[0]=0;		/* twice again?? */
 
 	for(i=0;i<ret;i++)	/* remove CF and LF characters */
 		if(buffer[i] == '\r' || buffer[i] == '\n')
 			buffer[i]='*';
 	
-	yubilog(LOG,"Client request",buffer,requestno);
+	//yubilog(LOG,"Client request",buffer,cli->fd);
 
 	/* Check here for Yubikey OTP request */
-	if(!strncmp(buffer,"GET /wsapi/2.0/verify",21))
+	if(cli->mode==EV_VAL_OTP)
 	{
-		fstr = (char *)malloc((size_t)BUFSIZE);
-		yubikey = (struct Yubikey *)calloc(1,sizeof(struct Yubikey));
-		null_terminate(buffer);
-		tokens = tokenize(buffer); /* Tokenize the buffer and retrieve main attributes */
-		premod_otp = tokens->otp;  /* Before modhex decoding OTP string */
-		
-		if(premod_otp == NULL)
-			yubilog(WARNING,"Only simple verification operations supported for OTP",0,fd);
-		
-		otp = modhex2hex(premod_otp);
-		id = tokens->id;
-		nonce = tokens->nonce;
-		otp = aes128ecb_decrypt(otp,premod_otp,private_id); /* otp is now decrypted if it is valid and if user exists */
-		
-		if(otp == NULL || strlen(premod_otp) < OTP_TOKEN || strlen(premod_otp) > OTP_TOKEN) 
-			result = 1;
-		else
-			result = validate_all(otp,premod_otp,private_id,id,yubikey,METHOD_OTP,NULL); /*validate*/
-		
-		t = time(NULL);
-		tmp = localtime(&t);
-		strftime(validation_date,20,"%Y-%m-%dT%H:%M:%S",tmp);
-		if(nonce!=NULL)
-			if(strlen(nonce) < 16 || strlen(nonce) > 40)
-				nonce = NULL;
-		
-		if(id!=NULL)
-		{	
-			hmac = create_hmac(premod_otp,status[result],validation_date,id,nonce); /* Add nonce when computing hash */
-			if(hmac==NULL)
-				result=NO_SUCH_CLIENT;
-		}
-		if(tokens->timestamp!=NULL)
-		{	
-			if(atoi(tokens->timestamp) == 1)
+		if (revents & EV_WRITE)
+		{
+			fstr = (char *)malloc((size_t)BUFSIZE);
+			private_id = (char *)malloc((size_t)PRIVATE_ID_SIZE);
+			yubikey = (struct Yubikey *)calloc(1,sizeof(struct Yubikey));
+			null_terminate(buffer);
+			tokens = tokenize(buffer); /* Tokenize the buffer and retrieve main attributes */
+			char *otp_n = NULL;
+			/* tokens->otp --------->  Before modhex decoding OTP string */
+			if(tokens->otp != NULL)
 			{
-				sessioncounter = yubikey->session_counter;
-				sessionuse = yubikey->session_token_counter;
-				session_ts = 1;
+				otp=(char *)malloc((size_t)OTP_TOKEN);
+				modhex2hex(otp,tokens->otp);
+				otp_n = aes128ecb_decrypt(otp,tokens->otp,private_id); /* otp is now decrypted if it is valid and if user exists */
+				if(otp != NULL) free(otp);
 			}
+			else if(tokens->otp == NULL)
+			{
+				yubilog(LOG,"Only simple verification operations supported for OTP, OTP is NULL",0,cli->fd);
+			}
+			if(otp_n == NULL || strlen(tokens->otp) < OTP_TOKEN || strlen(tokens->otp) > OTP_TOKEN) 
+				result = BAD_OTP;
+			else
+				result = validate_all(otp_n,tokens->otp,private_id,tokens->id,yubikey,METHOD_OTP,NULL); /*validate*/
+			
+			t = time(NULL);
+			ftime(&tp);
+			tmp = localtime(&t);
+			strftime(datetmp,20,"%FT%T",tmp);
+			sprintf(validation_date,"%s.%.3dZ",datetmp,tp.millitm);
+			if(tokens->nonce != NULL)
+				if(strlen(tokens->nonce) < 16 || strlen(tokens->nonce) > 40)
+					tokens->nonce = NULL;
+			
+			if(tokens->id!=NULL)
+			{
+				hmac = create_hmac(tokens->otp,status[result],validation_date,tokens->id,tokens->nonce); /* Add nonce when computing hash */
+				if(hmac==NULL)
+					result=NO_SUCH_CLIENT;
+			}
+			if(tokens->timestamp!=NULL)
+			{	
+				if(atoi(tokens->timestamp) == 1)
+				{
+					sessioncounter = yubikey->session_counter;
+					sessionuse = yubikey->session_token_counter;
+					session_ts = 1;
+				}
+			}
+			if(tokens->sl!=NULL)
+			{
+				sl = atoi(tokens->sl);
+			}
+			(void)sprintf(fstr,"HTTP/1.0 200 OK\r\nContent-Type: %s\r\n\r\n", "text/plain");
+			(void)write(cli->fd,fstr,strlen(fstr));
+			if(session_ts && result==OK)
+			{
+				(void)sprintf(fstr,"h=%s\r\nt=%s\r\notp=%s\r\nnonce=%s\r\nsl=%d\r\ntimestamp=%d\r\nsessioncounter=%d\r\nsessionuse=%d\r\nstatus=%s\r\n",
+						hmac!=NULL ? hmac : " ",validation_date,tokens->otp,tokens->nonce!=NULL? tokens->nonce : " ",sl!=-1? sl:100,yubikey->timestamp,sessioncounter,sessionuse, status[result]);
+			}
+			else{
+				(void)sprintf(fstr,"h=%s\r\nt=%s\r\notp=%s\r\nnonce=%s\r\nsl=%d\r\nstatus=%s\r\n",
+						hmac!=NULL ? hmac : " ",validation_date,tokens->otp,tokens->nonce!=NULL? tokens->nonce : " ",sl!=-1? sl:100 ,status[result]);
+			}
+			(void)write(cli->fd,fstr,strlen(fstr));
+			/* Update database internal counter and timestamp */
+			/* We are doing it here because we don't want the user to experience a small response delay */
+			if(result==OK) sqlite3_updatecounter(tokens->otp,yubikey->counter,yubikey->timestamp); 
+			free(fstr);
+			free(private_id);
+			free(yubikey);
+			if(hmac) free(hmac);
+			if(otp_n) free(otp_n);
+			free(tokens);
+			ev_io_stop(EV_A_ w); /* Preserve SQLite3 counter update */
 		}
-		if(tokens->sl!=NULL)
-		{
-			sl = atoi(tokens->sl);
-		}
-		(void)sprintf(fstr,"HTTP/1.0 200 OK\r\nContent-Type: %s\r\n\r\n", "text/plain");
-		(void)write(fd,fstr,strlen(fstr));
-		if(session_ts && result==OK)
-		{
-			(void)sprintf(fstr,"h=%s\r\nt=%s\r\notp=%s\r\nnonce=%s\r\nsl=%d\r\ntimestamp=%d\r\nsessioncounter=%d\r\nsessionuse=%d\r\nstatus=%s\r\n",
-					hmac!=NULL ? hmac : " ",validation_date,premod_otp,nonce!=NULL? nonce : " ",sl!=-1? sl:100,yubikey->timestamp,sessioncounter,sessionuse, status[result]);
-		}
-		else{
-			(void)sprintf(fstr,"h=%s\r\nt=%s\r\notp=%s\r\nnonce=%s\r\nsl=%d\r\nstatus=%s\r\n",
-					hmac!=NULL ? hmac : " ",validation_date,premod_otp,nonce!=NULL? nonce : " ",sl!=-1? sl:100 ,status[result]);
-		}
-		(void)write(fd,fstr,strlen(fstr));
-		/* Update database internal counter and timestamp */
-		/* We are doing it here because we don't want the user to experience a small response delay */
-		if(result==OK) sqlite3_updatecounter(premod_otp,yubikey->counter,yubikey->timestamp); 
-		free(fstr);
-		free(private_id);
-		free(yubikey);
-		free(tokens);
-		if(!hmac) free(hmac);
-	}else if(!strncmp(buffer,"GET /wsapi/2.0/oauthverify",26)) /* Check here for OAUTH Yubikey request */
+	}else if(cli->mode==EV_VAL_OATH) /* Check here for OATH Yubikey request */
 	{
-		fstr = (char *)malloc((size_t)BUFSIZE);
-     		null_terminate(buffer);
-                oauth_tokens = oauth_tokenize(buffer); /* Tokenize the buffer and retrieve main attributes */
-		t = time(NULL);
-		tmp = localtime(&t);
-		strftime(validation_date,20,"%Y-%m-%dT%H:%M:%S",tmp);
-
-		if(oauth_tokens->otp != NULL && ( strlen(oauth_tokens->otp) == 18 || strlen(oauth_tokens->otp) == 20  ) )
-			result = validate_all(oauth_tokens->otp,NULL,NULL,NULL,NULL,METHOD_OAUTH,&oauth_tokens->otp);
-		else
-			result = 1;
-		(void)sprintf(fstr,"HTTP/1.0 200 OK\r\nContent-Type: %s\r\n\r\n", "text/plain");
-		(void)write(fd,fstr,strlen(fstr));
-		(void)sprintf(fstr,"h=%s\r\nt=%s\r\notp=%s\r\nstatus=%s\r\n",
-					" ",validation_date,oauth_tokens->otp,status[result]);
+		if (revents & EV_WRITE)
+		{
+			fstr = (char *)malloc((size_t)BUFSIZE);
+	     		null_terminate(buffer);
+	                oath_tokens = oath_tokenize(buffer); /* Tokenize the buffer and retrieve main attributes */
+			t = time(NULL);
+			ftime(&tp);
+			tmp = localtime(&t);
+			strftime(datetmp,20,"%FT%T",tmp);
+			sprintf(validation_date,"%s.%.3dZ",datetmp,tp.millitm);
+			if(oath_tokens->otp != NULL && ( strlen(oath_tokens->otp) == 18 || strlen(oath_tokens->otp) == 20  ) )
+				result = validate_all(oath_tokens->otp,NULL,NULL,NULL,NULL,METHOD_OATH,&oath_tokens->otp);
+			else
+				result = 1;
+			(void)sprintf(fstr,"HTTP/1.0 200 OK\r\nContent-Type: %s\r\n\r\n", "text/plain");
+			(void)write(cli->fd,fstr,strlen(fstr));
+			(void)sprintf(fstr,"h=%s\r\nt=%s\r\notp=%s\r\nstatus=%s\r\n",
+						" ",validation_date,oath_tokens->otp,status[result]);
 		
-		(void)write(fd,fstr,strlen(fstr));
-		if(result==OK) sqlite3_oauth_updatecounter(oauth_tokens->otp,oauth_tokens->counter);
-		free(fstr);
-		free(oauth_tokens);
-	}else if(!strncmp(buffer,"GET / HTTP/1.1",14)) /* Default request */
+			(void)write(cli->fd,fstr,strlen(fstr));
+			if(result==OK) sqlite3_oath_updatecounter(oath_tokens->otp,oath_tokens->counter);
+			free(fstr);
+			free(oath_tokens);
+			ev_io_stop(EV_A_ w);
+		}
+	}else if(cli->mode=EV_DEFAULT_PAGE) /* Default HTTP/1.0 request */
 	{
-		(void)sprintf(buffer,"HTTP/1.0 200 OK\r\nContent-Type: %s\r\n\r\n", "text/html");
-		(void)write(fd,buffer,strlen(buffer));
-		fstr="<html>Yubico Yubikey:<br><form action='/wsapi/2.0/verify' method='GET'> \
-		     <input type='text' name='otp'><br><input type='submit'></form><br>OAUTH/HOTP tokens:<br> \
-		     <form action='/wsapi/2.0/oauthverify' method='GET'>\
-		     <input type='text' name='otp'><br><input type='submit'></form></html>\r\n";
-		(void)write(fd,fstr,strlen(fstr));
-	}else if(!strncmp(buffer,"GET / HTTP/1.0",14)) /* Default request */
+		if (revents & EV_WRITE)
+		{
+			(void)sprintf(buffer,"HTTP/1.0 200 OK\r\nContent-Type: %s\r\n\r\n", "text/html");
+			(void)write(cli->fd,buffer,strlen(buffer));
+			(void)sprintf(buffer,"<html>Yubico Yubikey:<br><form action='/wsapi/2.0/verify' method='GET'> \
+			     <input type='text' name='otp'><br><input type='submit'></form><br>OATH/HOTP tokens:<br> \
+			     <form action='/wsapi/2.0/oauthverify' method='GET'>\
+			     <input type='text' name='otp'><br><input type='submit'></form></html>\r\n");
+			(void)write(cli->fd,buffer,strlen(buffer));
+			ev_io_stop(EV_A_ w);
+		}
+	}else if(cli->mode==EV_DEFAULT_PAGE) /* Default HTTP/1.1 request */
 	{
-		(void)sprintf(fstr,"HTTP/1.0 200 OK\r\nContent-Type: %s\r\n\r\n", "text/html");
-		(void)write(fd,fstr,strlen(buffer));
-		fstr="<html>Yubico Yubikey:<br><form action='/wsapi/2.0/verify' method='GET'> \
-		     <input type='text' name='otp'><br><input type='submit'></form><br>OAUTH/HOTP tokens:<br> \
-		     <form action='/wsapi/2.0/oauthverify' method='GET'> \
-		     <input type='text' name='otp'><br><input type='submit'></form></html>\r\n";
-                (void)write(fd,fstr,strlen(fstr));
+		if (revents & EV_WRITE)
+		{
+			(void)sprintf(buffer,"HTTP/1.0 200 OK\r\nContent-Type: %s\r\n\r\n", "text/html");
+			(void)write(cli->fd,buffer,strlen(buffer));
+			(void)sprintf(buffer,"<html>Yubico Yubikey:<br><form action='/wsapi/2.0/verify' method='GET'> \
+			     <input type='text' name='otp'><br><input type='submit'></form><br>OATH/HOTP tokens:<br> \
+			     <form action='/wsapi/2.0/oauthverify' method='GET'> \
+			     <input type='text' name='otp'><br><input type='submit'></form></html>\r\n");
+	                (void)write(cli->fd,buffer,strlen(buffer));
+			ev_io_stop(EV_A_ w);
+		}
 	}else
 	{ /*Maybe we should redirect this back to the default request???*/
-		yubilog(WARNING,"Only simple verification operations supported for OTP/OAUTH",buffer,fd);
+		if (revents & EV_WRITE)
+		{
+			//yubilog(WARNING,"Only simple verification operations supported for OTP/OATH",buffer,cli->fd);
+			ev_io_stop(EV_A_ w);
+		}
 	}
-	//sleep(1);	/* to allow socket to drain */
-	close(fd);
-	exit(1);
+read_error:
+	close(cli->fd); /* Don't let client to drain */
+	free(buffer);
+	free(cli);
 }
 
 void usage()
 {
 	fprintf(stderr,
-			"Simple and lightweight Yubikey OTP and HOTP/OAUTH validation server.\n"
+			"Simple and lightweight Yubikey OTP and HOTP/OATH validation server.\n"
 		        "Version " VERSION ". Written and copyrights by Nanakos Chrysostomos.\n"
 			"THIS SOFTWARE COMES WITH ABSOLUTELY NO WARRANTY! USE AT YOUR OWN RISK!\n"	
 			"\nUsage: yubiserver [options] \n\n"
 			"Options supported:\n"
-			"   --version or -V		Print version information\n"
-			"   --help or -h			Print this help screen\n"
+			"   --version  or -V		Print version information\n"
+			"   --help     or -h		Print this help screen\n"
 			"   --database or -d		Use this SQLite3 database file\n"
-			"   --port or -p			Port to bind the server. Default port is 8000\n"
-			"   --logfile or -l		Use this as logfile. Default is '" YUBISERVER_LOG_PATH "'\n"
+			"   --port     or -p		Port to bind the server. Default port is 8000\n"
+			"   --logfile  or -l		Use this as logfile. Default is '" YUBISERVER_LOG_PATH "'\n"
 			"This version of yubiserver has been configured with '" SQLITE3_DB_PATH "' as its default\n"
 			"SQLite3 database file.\n");
 }
 
+int setnonblock(int fd)
+{
+    int flags;
+    flags = fcntl(fd, F_GETFL);
+    if (flags < 0)
+            return flags;
+    flags |= O_NONBLOCK;
+    if (fcntl(fd, F_SETFL, flags) < 0) 
+            return -1;
+    return 0;
+}
+
+static void read_callback(struct ev_loop *loop, struct ev_io *w, int revents)
+{ 
+	
+	struct ev_client *cli= ((struct ev_client*) (((char*)w) - offsetof(struct ev_client,ev_read)));
+	long r=0;
+	char rbuff[BUFSIZE+1];
+	if (revents & EV_READ){
+		r=read(cli->fd,&rbuff,BUFSIZE);
+		rbuff[r]='\0';
+		cli->buffer = rbuff;
+		cli->ret = r;
+		if(!strncmp(rbuff,"GET / HTTP/1.0",14))
+		{
+			cli->mode = EV_DEFAULT_PAGE;
+		}else if(!strncmp(rbuff,"GET / HTTP/1.1",14))
+		{
+			cli->mode = EV_DEFAULT_PAGE;
+		}else if(!strncmp(rbuff,"GET /wsapi/2.0/verify",21))
+	        {
+			cli->mode = EV_VAL_OTP;
+		}else if(!strncmp(rbuff,"GET /wsapi/2.0/oauthverify",26))
+		{
+			cli->mode = EV_VAL_OATH;
+		}
+
+	}
+	ev_io_stop(EV_A_ w);
+	ev_io_init(&cli->ev_write,write_callback,cli->fd,EV_WRITE);
+	ev_io_start(loop,&cli->ev_write);
+	
+}
+
+static void accept_callback(struct ev_loop *loop, struct ev_io *w, int revents)
+{
+	int client_fd;
+	struct ev_client *client;
+	struct sockaddr_in client_addr;
+	socklen_t client_len = sizeof(client_addr);
+	client_fd = accept(w->fd, (struct sockaddr *)&client_addr, &client_len);
+        if (client_fd == -1) {
+                return;
+        }
+   	 
+	client = calloc(1,sizeof(*client));
+	client->fd=client_fd;
+	if (setnonblock(client->fd) < 0)
+                yubilog(LOG, "failed to set client socket to non-blocking",0,0);
+	ev_io_init(&client->ev_read,read_callback,client->fd,EV_READ);
+	ev_io_start(loop,&client->ev_read);
+}
+
+//#define EVFLAG_FORKCHECK 1
+
 int main(int argc, char **argv)
 {
-	int i, port, pid, listenfd, socketfd, requestno;
-	socklen_t length;
+	int i, port, listenfd ;
 	char portStr[6];
-	static struct sockaddr_in cli_addr; /* static = initialised to zeros */
+	int reuseaddr_on = 1;
 	static struct sockaddr_in serv_addr; /* static = initialised to zeros */
+	
+	/* EV */
+	ev_io *ev_accept = (ev_io *)malloc(sizeof(ev_io));
+	struct ev_loop *loop = ev_default_loop (0);
+
 	struct option long_options[] = 
 	{
 		/*Takes no parameters*/
@@ -1037,10 +1216,14 @@ int main(int argc, char **argv)
 		(void)close(i);		/* close open files */
 	
 	(void)setpgrp();		/* break away from process group */
-	/* setup the network socket */
 
+	/* setup the network socket */
 	if((listenfd = socket(AF_INET, SOCK_STREAM,0)) <0)
 		yubilog(ERROR, "system call","socket",0);
+
+	if (setsockopt(listenfd, SOL_SOCKET, SO_REUSEADDR, &reuseaddr_on,sizeof(reuseaddr_on)) == -1)
+            yubilog(ERROR, "setsockopt failed",NULL,0);
+
 	sprintf(portStr,"%d",port);
 
 	if(port < 1024 || port > 65534)
@@ -1056,25 +1239,14 @@ int main(int argc, char **argv)
 
 	if( listen(listenfd,64) <0)
 		yubilog(ERROR,"system call","listen",0);
+	
+	if (setnonblock(listenfd) < 0)
+            yubilog(ERROR, "cannot set server listening socket to non-blocking",0,0);
 
+	ev_io_init(ev_accept,accept_callback,listenfd,EV_READ);
+	ev_io_start(loop,ev_accept);
+	ev_loop_fork(loop);
+	ev_loop (loop, 0);
 
-	for(requestno=1; ;requestno++) 
-	{
-		length = sizeof(cli_addr);
-
-		if((socketfd = accept(listenfd, (struct sockaddr *)&cli_addr,&length)) < 0)
-			yubilog(ERROR,"system call","accept",0);
-
-		if((pid = fork()) < 0) {
-			yubilog(ERROR,"system call","fork",0);
-		}
-		else {
-			if(pid == 0) { 	/* child */
-				(void)close(listenfd);
-				authentication_handle(socketfd,requestno); /* never returns */
-			} else { 	/* parent */
-				(void)close(socketfd);
-			}
-		}
-	}
+	return 0;
 }
